import logging
import re
from typing import Dict
from pygls.server import LanguageServer
from lsprotocol import types
from pygls.workspace import TextDocument
from pygls import uris
from couleuvre.ast_parser.vyper_ast import BaseNode
from couleuvre.cli import start_server
from couleuvre.logger_setup import setup_logging
from couleuvre import utils
from couleuvre.parser.parse import Module, parse_module
from couleuvre.features.symbols_visitor import get_document_symbols

logger = logging.getLogger("couleuvre")


class VyperLanguageServer(LanguageServer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.modules: Dict = {}
        self.logger = setup_logging(self)  # Inject logging on init
        self.logger.info("Vyper Language Server starting...")
        self.default_version = None

    def parse(self, doc: TextDocument, workspace_path=None):
        self.modules[doc.uri] = parse_module(
            doc.path,
            default_version=self.default_version,
            workspace_path=workspace_path,
        )
        if not self.default_version:
            self.default_version = self.modules[doc.uri].version
        self.logger.info("Module: %s", self.modules)

    def get_module(self, doc: TextDocument, workspace_path=None):
        self.logger.info(f"get module: {doc.uri}")
        if doc.uri not in self.modules:
            self.modules[doc.uri] = parse_module(
                doc.path,
                default_version=self.default_version,
                workspace_path=workspace_path,
            )
            if not self.default_version:
                self.default_version = self.modules[doc.uri].version
        return self.modules[doc.uri]


server = VyperLanguageServer("couleuvre", "v0.0.1")


@server.feature(types.TEXT_DOCUMENT_DID_OPEN)
def did_open(ls: VyperLanguageServer, params: types.DidOpenTextDocumentParams):
    """Parse each document when it is opened"""
    ls.logger.info(f"Document opened: {params.text_document.uri}")
    doc = ls.workspace.get_text_document(params.text_document.uri)
    ls.parse(doc, workspace_path=ls.workspace.root_path)


@server.feature(types.TEXT_DOCUMENT_DID_CHANGE)
def did_change(ls: VyperLanguageServer, params: types.DidChangeTextDocumentParams):
    """Parse each document when it is changed"""
    ls.logger.info(f"Document changed: {params.text_document.uri}")
    doc = ls.workspace.get_text_document(params.text_document.uri)
    ls.parse(doc, workspace_path=ls.workspace.root_path)


@server.feature(types.TEXT_DOCUMENT_DOCUMENT_SYMBOL)
def document_symbol(ls: VyperLanguageServer, params: types.DocumentSymbolParams):
    """Return all the symbols defined in the given document."""
    ls.logger.info(f"Document symbol requested: {params.text_document.uri}")
    doc = ls.workspace.get_text_document(params.text_document.uri)
    symbols = get_document_symbols(
        ls.get_module(doc, workspace_path=ls.workspace.root_path)
    )
    return symbols


@server.feature(types.WORKSPACE_SYMBOL)
def workspace_symbol(ls: VyperLanguageServer, params: types.WorkspaceSymbolParams):
    """Return all the symbols defined in the given document."""
    return


def goto_in_module(doc: TextDocument, module: Module, chain: list, external=False):
    if external:
        res = module.external_namespace()
    else:
        res = module.namespace
    while chain and res:
        res = res.get(chain.pop(0), None)
    if res is not None and isinstance(res, BaseNode):
        return types.Location(uri=doc.uri, range=utils.range_from_node(res))


@server.feature(types.TEXT_DOCUMENT_DEFINITION)
def goto_definition(ls: VyperLanguageServer, params: types.DefinitionParams):
    """Jump to an object's definition."""
    ls.logger.info(f"Definition requested: {params.text_document.uri}")
    doc = ls.workspace.get_text_document(params.text_document.uri)
    module = ls.get_module(doc, workspace_path=ls.workspace.root_path)

    try:
        attribute_word = doc.word_at_position(
            params.position, re.compile(r"[A-Za-z_0-9]+(?:\.[A-Za-z_0-9]+)*$")
        )
    except IndexError:
        return
    if not attribute_word:
        return

    # Local lookup
    target = goto_in_module(doc, module, attribute_word.split("."))
    if target:
        return target

    # Imported module lookup
    parts = attribute_word.split(".")
    root_name, remainder = parts[0], parts[1:]

    if root_name in module.imports.keys():
        resolved_path = module.imports[root_name]
        resolved_uri = uris.from_fs_path(resolved_path)
        if not resolved_uri:
            return
        resolved_doc = ls.workspace.get_text_document(resolved_uri)
        resolved_module = ls.get_module(
            resolved_doc, workspace_path=ls.workspace.root_path
        )
        if not remainder:
            return utils.location_from_start(resolved_doc.uri)

        return goto_in_module(resolved_doc, resolved_module, remainder, external=True)


# TODO signature help
# TODO implementation
# TODO hover
# TODO reference
# TODO definition
# TODO declaration
# TODO completion


def main():
    # TODO: replace by pygls.cli once it is released
    start_server(server)



import os
from pathlib import Path
import subprocess
import json
import sys
from couleuvre.ast_parser.vyper_wrapper import ensure_vyper_version
from couleuvre.ast_parser.vyper_ast import BaseNode, AST_CLASS_MAP
import logging
from typing import Any, Dict, Optional
from packaging.version import Version
from couleuvre import utils  # <--- added

logger = logging.getLogger("vyper-lsp")


def _obtain_sys_path(python_bin) -> list[str]:
    """
    Obtain all the system paths in which the compiler would
    normally look for modules. This allows to lint vyper files
    that import a module that is installed as a virtual environment
    dependency.
    """

    command = [python_bin, "-c", "import sys; print(sys.path)"]

    process = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
    )
    stdout, stderr = process.communicate()

    paths = stdout.strip().strip("[]").replace("'", "").split(", ")
    return paths


def get_search_paths(
    python_bin: Optional[str] = None,
    include_sys_path: bool = True,
) -> list[str]:
    """
    Build the search path list passed to Vyper's FilesystemInputBundle.

    - If include_sys_path is False, return an empty list (no search paths).
    - If include_sys_path is True, obtain sys.path from the provided python_bin
      (intended for the couleuvre-managed env) and include "." for workspace resolution.
    """
    search_paths: list[str] = []

    if include_sys_path and python_bin:
        search_paths = _obtain_sys_path(python_bin)
        if "." not in search_paths:
            search_paths.append(".")

    logger.info(f"Final search paths: {search_paths}")
    return search_paths


def get_script(file_path: str, vyper_version: str, search_paths: list[str]) -> str:
    if Version(vyper_version) >= Version("0.4.1"):
        pass
        return f"""
import copy
import json
from vyper.semantics.analysis.imports import resolve_imports
from packaging.version import Version
from pathlib import Path
from vyper.compiler import CompilerData
from vyper.compiler.input_bundle import FilesystemInputBundle
search_paths = {repr(search_paths)}
search_paths = [Path(p) for p in search_paths]
input_bundle = FilesystemInputBundle(search_paths)
file = input_bundle.load_file({repr(file_path)})
module = CompilerData(file, input_bundle).vyper_module
with input_bundle.search_path(Path(module.resolved_path).parent):
    imports = resolve_imports(module, input_bundle)

print(json.dumps(module.to_dict()))
"""
    else:
        content = Path(file_path).read_text()
        return f"""
import json
from vyper.compiler import CompilerData
data = CompilerData({repr(content)}).vyper_module
print(json.dumps(data.to_dict()))
"""


def get_json_ast(
    path: str, vyper_version: str, workspace_path: Optional[str] = None
) -> BaseNode:
    # Decide which Python to use:
    use_local_env = False
    try:
        use_local_env = utils.get_installed_vyper_version() == Version(vyper_version)
    except Exception:
        use_local_env = False

    if use_local_env:
        # Use current interpreter; no search paths
        python_bin = sys.executable
        logger.info(f"Using local environment's vyper {vyper_version}")
    else:
        # Use couleuvre-managed venv and include its sys.path as search paths
        venv_path = ensure_vyper_version(vyper_version)
        python_bin = os.path.join(Path(venv_path), "bin", "python")
        logger.info(f"Using couleuvre-managed vyper {vyper_version} at {python_bin}")

    search_paths = get_search_paths(
        python_bin=python_bin,
        include_sys_path=True,
    )
    script = get_script(path, vyper_version, search_paths)

    result = subprocess.run(
        [python_bin, "-c", script],
        shell=True,
        input=script,
        capture_output=True,
        text=True,
        cwd=workspace_path,
    )

    if result.returncode != 0:
        logger.error(
            f"Failed to get AST for Vyper {vyper_version}: subprocess error: {result.stderr.strip()}"
        )
    else:
        logger.info(f"Got Vyper AST from Vyper {vyper_version}")

    parsed_json = json.loads(result.stdout)
    lsp_ast = _from_vyper_json_ast(parsed_json)

    logger.info("Parsed Vyper AST to LSP AST")
    return lsp_ast


# === Converter ===


def _from_vyper_json_ast(
    ast_dict: Dict[str, Any], parent: Optional[BaseNode] = None
) -> BaseNode:
    ast_type = ast_dict["ast_type"]
    if ast_type == "List":
        ast_type = "ListNode"
    elif ast_type == "Tuple":
        ast_type = "TupleNode"
    elif ast_type == "Dict":
        ast_type = "DictNode"
    elif ast_type == "EnumDef":
        ast_type = "FlagDef"
    cls = AST_CLASS_MAP.get(ast_type, BaseNode)
    cls_fields = cls.__dataclass_fields__

    kwargs = {"ast_type": ast_type}

    for key, value in ast_dict.items():
        if key not in cls_fields:
            logger.error(
                f"Key '{str(key)}' not found in {str(cls.__name__)} dataclass fields"
            )
            continue
        if isinstance(value, list):
            children = []
            for item in value:
                if isinstance(item, dict) and "ast_type" in item:
                    child = _from_vyper_json_ast(item)
                    child.parent = None  # Temporary
                    children.append(child)
                else:
                    children.append(item)
            kwargs[key] = children
        elif isinstance(value, dict) and "ast_type" in value:
            child = _from_vyper_json_ast(value)
            child.parent = None  # Temporary
            kwargs[key] = child
        else:
            kwargs[key] = value
    # TODO sus type ignore
    node = cls(**kwargs)  # type: ignore
    node.parent = parent

    # Now update children with the correct parent reference
    for key, value in kwargs.items():
        if isinstance(value, BaseNode):
            value.parent = node
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, BaseNode):
                    item.parent = node

    return node



import logging
from typing import List, Tuple

from lsprotocol import types
from lsprotocol.types import SymbolKind

from couleuvre.ast_parser import vyper_ast
from couleuvre.parser.parse import Module
from couleuvre.utils import range_from_node

logger = logging.getLogger("vyper-lsp")


class VyperNodeVisitorBase:
    ignored_types: Tuple = ()
    scope_name = ""

    def visit(self, node, *args):
        if isinstance(node, self.ignored_types):
            return []
        node_type = type(node).__name__
        visitor_fn = getattr(self, f"visit_{node_type}", None)
        if visitor_fn is None:
            return []
        return visitor_fn(node, *args)


def get_document_symbols(module: Module) -> List[types.DocumentSymbol]:
    visitor = SymbolVisitor()
    return visitor.visit(module.ast)


def get_symbol(node, name, kind, children=None):
    if children is None:
        children = []
    return types.DocumentSymbol(
        name=name,
        kind=kind,
        range=range_from_node(node),
        selection_range=range_from_node(node),
        children=children,
    )


class SymbolVisitor(VyperNodeVisitorBase):
    def visit_Module(self, node):
        symbols = []
        for child in node.body:
            symbols += self.visit(child)
        return symbols

    def visit_VariableDecl(self, node):
        name = node.target.id
        kind = SymbolKind.Variable
        if node.is_constant or node.is_immutable:
            kind = SymbolKind.Constant
        return [get_symbol(node, name, kind, [])]

    def visit_FlagDef(self, node):
        children = []
        for child in node.body:
            assert isinstance(child, vyper_ast.Expr)
            children += self.visit(child.value, SymbolKind.EnumMember)
        return [get_symbol(node, node.name, SymbolKind.Enum, children)]

    def visit_EventDef(self, node):
        return self._visit_struct_like(node, SymbolKind.Event)

    def visit_StructDef(self, node):
        return self._visit_struct_like(node, SymbolKind.Struct)

    def _visit_struct_like(self, node, symbol):
        children = []
        for child in node.body:
            children += self.visit(child, SymbolKind.Field)
        return [get_symbol(node, node.name, symbol, children)]

    def visit_FunctionDef(self, node, kind=SymbolKind.Function):
        children = []
        children += self.visit(node.args)
        for child in node.body:
            children += self.visit(child)
        name = node.name
        return [get_symbol(node, name, kind, children)]

    def visit_InterfaceDef(self, node):
        children = []
        for child in node.body:
            children += self.visit(child, SymbolKind.Method)
        return [get_symbol(node, node.name, SymbolKind.Interface, children)]

    def visit_arguments(self, node):
        args = []
        for arg in node.args:
            args += self.visit(arg)
        return args

    def visit_arg(self, node):
        return [get_symbol(node, node.arg, SymbolKind.Variable)]

    def visit_AnnAssign(self, node, kind=None):
        if not kind:
            # Old vyper version would have variable declaration as AnnAssign
            if isinstance(node.parent, vyper_ast.Module):
                return [get_symbol(node, node.target.id, SymbolKind.Variable)]
        # assert kind in (SymbolKind.Field,) removed because of default args most likely.
        if isinstance(node.parent, (vyper_ast.EventDef, vyper_ast.StructDef)):
            return [get_symbol(node, node.target.id, SymbolKind.Field)]
        return []

    def visit_For(self, node):
        symbols = []
        for child in node.body:
            symbols += self.visit(child)
        symbols += self.visit(node.target)  # AnnAssign
        return symbols

    def visit_If(self, node):
        symbols = []
        for child in node.body:
            symbols += self.visit(child)
        for child in node.orelse:
            symbols += self.visit(child)
        return symbols

    def visit_Name(self, node, kind=SymbolKind.Variable):
        if isinstance(node.parent.parent, vyper_ast.FlagDef):
            assert kind == SymbolKind.EnumMember
            return [get_symbol(node, node.id, kind)]
        return []



############################################################################
# Copyright(c) Open Law Library. All rights reserved.                      #
# See ThirdPartyNotices.txt in the project root for additional notices.    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License")           #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#     http: // www.apache.org/licenses/LICENSE-2.0                         #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
############################################################################
"""A simple cli wrapper for pygls servers."""

from __future__ import annotations

import argparse


def start_server(server, args: list[str] | None = None):
    """A helper function that implements a simple cli wrapper for a pygls server
    allowing the user to select between the supported transports."""

    name = type(server).__name__
    parser = argparse.ArgumentParser(description=f"start a {name} instance")
    parser.add_argument("--tcp", action="store_true", help="start a TCP server")
    parser.add_argument("--ws", action="store_true", help="start a WebSocket server")
    parser.add_argument("--host", default="127.0.0.1", help="bind to this address")
    parser.add_argument("--port", type=int, default=8888, help="bind to this port")

    arguments = parser.parse_args(args)

    if arguments.tcp:
        server.start_tcp(arguments.host, arguments.port)
    elif arguments.ws:
        server.start_ws(arguments.host, arguments.port)
    else:
        server.start_io()



import logging
from importlib.metadata import version
from lsprotocol.types import Position, Range, Location
from packaging.version import Version
from couleuvre.ast_parser.vyper_ast import BaseNode

logger = logging.getLogger("couleuvre")


def get_installed_vyper_version():
    return Version(version("vyper"))


def range_from_node(node: BaseNode) -> Range:
    return Range(
        start=Position(line=node.lineno - 1, character=node.col_offset),
        end=Position(line=node.end_lineno - 1, character=node.end_col_offset),
    )


def range_from_start() -> Range:
    return Range(
        start=Position(line=0, character=0),
        end=Position(line=0, character=0),
    )


def location_from_start(uri: str) -> Location:
    return Location(uri=uri, range=range_from_start())



import logging
from pygls.server import LanguageServer


class LspLogHandler(logging.Handler):
    """Custom log handler that sends logs to the LSP client."""

    def __init__(self, ls: LanguageServer):
        super().__init__()
        self.ls = ls

    def emit(self, record):
        try:
            message = self.format(record)
            # Avoid sending logs if the server isn't fully initialized
            if self.ls and hasattr(self.ls, "show_message_log"):
                self.ls.show_message_log(message)
        except Exception:
            self.handleError(record)


def setup_logging(ls: LanguageServer, level=logging.INFO) -> logging.Logger:
    """Configures logging for the LSP server."""
    logger = logging.getLogger("vyper-lsp")
    logger.setLevel(level)

    # Prevent duplicate log handlers in case of reload
    if not logger.hasHandlers():
        # Console handler (stdout for debugging)
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        console_formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        console_handler.setFormatter(console_formatter)

        # LSP log handler
        lsp_handler = LspLogHandler(ls)
        lsp_handler.setLevel(level)
        lsp_formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
        lsp_handler.setFormatter(lsp_formatter)

        logger.addHandler(console_handler)
        logger.addHandler(lsp_handler)

    return logger



import subprocess
from pathlib import Path
from packaging.version import Version

VYPER_BASE_DIR = Path.home() / ".couleuvre" / "venvs"


def _get_venv_path(version: str) -> Path:
    return VYPER_BASE_DIR / version


def _get_py_version_for_vy_version(vy_version: str) -> str:
    vy_ver = Version(vy_version)

    if vy_ver <= Version("0.2.7"):
        return "3.8"
    if vy_ver <= Version("0.3.2"):
        return "3.9"
    return "3.10"


def ensure_vyper_version(version: str) -> Path:
    """
    Ensure the specified vyper version is available in a uv-managed virtual environment.
    Returns the path to the virtual environment.
    """
    venv_path = _get_venv_path(version)

    if not venv_path.exists():
        print(f"[couleuvre] Creating uv env for vyper {version}...")

        py_version = _get_py_version_for_vy_version(version)
        # Create the environment and install vyper using uv
        # uv venv --python=3.x <venv_path>
        subprocess.run(
            f"uv venv --python {py_version} {venv_path}",
            shell=True,
            check=True,
        )
        if Version(version) <= Version("0.2.7"):
            subprocess.run(
                f"source {venv_path}/bin/activate && uv pip install --upgrade setuptools",
                shell=True,
                check=True,
            )
        # uv pip install vyper==<version>
        subprocess.run(
            f"source {venv_path}/bin/activate && uv pip install vyper=={version}",
            shell=True,
            check=True,
        )

    return venv_path



import logging
import re
from typing import Optional, Tuple
from couleuvre.ast_parser import vyper_ast
from couleuvre.ast_parser.ast_parser import get_json_ast
from pathlib import Path

logger = logging.getLogger("couleuvre")


def parse_module(
    path: str,
    default_version: Optional[str] = None,
    workspace_path: Optional[str] = None,
) -> "Module":
    pattern = r"#\s*(?:@version|pragma\s+version)\s*(?:[<>=!~^]*)\s*(\d+\.\d+\.\d+)"
    content = Path(path).read_text()
    match = re.search(pattern, content)
    if not match:
        if default_version is not None:
            version = default_version
        else:
            raise ValueError("Version not found in the content")
    else:
        version = match.group(1)
    assert version is not None
    vyper_module = get_json_ast(path, version, workspace_path=workspace_path)
    visitor = VyperAstVisitor(vyper_module, version)
    visitor.visit(vyper_module)
    return visitor.module


class Import:
    def __init__(self, module: Optional[str], name: str, alias: Optional[str] = None):
        self.module = module
        self.name = name
        self.alias = alias


class Module:
    def __init__(self, ast, vyper_version):
        self.version = vyper_version
        self.ast = ast
        self.namespace = {"self": {}}

        self.flags = set()
        self.functions = set()
        self.events = set()
        self.interfaces = set()
        self.structs = set()
        self.variables = set()
        self.imports = dict()

    def external_namespace(self):
        return {
            k: v for k, v in self.namespace.items() if k != "self"
        } | self.namespace["self"]


class VyperNodeVisitorBase:
    ignored_types: Tuple = ()
    scope_name = ""

    def visit(self, node, *args):
        if isinstance(node, self.ignored_types):
            return
        node_type = type(node).__name__
        visitor_fn = getattr(self, f"visit_{node_type}", None)
        if visitor_fn is None:
            logger.info(
                f"Unsupported syntax for {self.scope_name} namespace: {node_type}",
                node,
            )
            return
        visitor_fn(node, *args)


class VyperAstVisitor(VyperNodeVisitorBase):
    def __init__(self, node, vyper_version: str):
        self.module: Module = Module(node, vyper_version)

    def visit_Module(self, node):
        for child in node.body:
            self.visit(child)
        pass

    def visit_VariableDecl(self, node):
        self.module.variables.add(node)
        if node.is_constant or node.is_immutable:
            self.module.namespace[node.target.id] = node
        else:
            self.module.namespace["self"][node.target.id] = node

    def visit_FunctionDef(self, node):
        self.module.functions.add(node)
        self.module.namespace["self"][node.name] = node

    def visit_FlagDef(self, node):
        self.module.flags.add(node)
        self.module.namespace[node.name] = node

    def visit_EventDef(self, node):
        self.module.events.add(node)
        self.module.namespace[node.name] = node

    def visit_InterfaceDef(self, node):
        self.module.interfaces.add(node)
        self.module.namespace[node.name] = node

    def visit_StructDef(self, node):
        self.module.structs.add(node)
        self.module.namespace[node.name] = node

    def _handle_import(self, node):
        resolved_path = (
            node.import_info.get("resolved_path") if node.import_info else None
        )
        if resolved_path is None:
            return
        if node.alias:
            self.module.imports[node.alias] = resolved_path
        self.module.imports[node.name] = resolved_path

    def visit_Import(self, node):
        self._handle_import(node)

    def visit_ImportFrom(self, node):
        self._handle_import(node)

    def visit_ImplementsDecl(self, node):
        pass

    def visit_UsesDecl(self, node):
        pass

    def visit_InitializesDecl(self, node):
        pass

    def visit_ExportsDecl(self, node):
        pass

    # For backwards compatibility
    def visit_AnnAssign(self, node):
        if isinstance(node.parent, vyper_ast.Module):
            if node.annotation is not None and isinstance(
                node.annotation, vyper_ast.Call
            ):
                if (
                    node.annotation.func.id == "constant"
                    or node.annotation.func.id == "immutable"
                ):
                    self.module.namespace[node.target.id] = node
            self.module.namespace["self"][node.target.id] = node
